# -*- coding: utf-8 -*-
"""
Created on Fri Nov 14 18:17:20 2014

@author: proto
"""

import libsbml
from sbml2bngl import SBML2BNGL as SBML2BNGL
import structures
import atomizer.moleculeCreation as mc
import os
from subprocess import call
import tempfile
import sys

# sys.path.insert(0, '../utils/')
import consoleCommands
import readBNGXML
import argparse

bioqual = [
    "BQB_IS",
    "BQB_HAS_PART",
    "BQB_IS_PART_OF",
    "BQB_IS_VERSION_OF",
    "BQB_HAS_VERSION",
    "BQB_IS_HOMOLOG_TO",
    "BQB_IS_DESCRIBED_BY",
    "BQB_IS_ENCODED_BY",
    "BQB_ENCODES",
    "BQB_OCCURS_IN",
    "BQB_HAS_PROPERTY",
    "BQB_IS_PROPERTY_OF",
    "BQB_HAS_TAXON",
    "BQB_UNKNOWN",
]

modqual = ["BQM_IS", "BQM_IS_DESCRIBED_BY", "BQM_IS_DERIVED_FROM", "BQM_UNKNOWN"]

from collections import defaultdict
import re


class AnnotationExtractor:
    def __init__(self, fileName, speciesEquivalencesFlag=None):
        self.fileName = fileName
        reader = libsbml.SBMLReader()
        self.sct = None
        self.database = None
        self.sbmlDocument = reader.readSBMLFromFile(fileName)

    def standardizeName(self, name):
        """
        Remove stuff not used by bngl
        """
        name2 = name

        sbml2BnglTranslationDict = {
            "^": "",
            "'": "",
            "*": "m",
            " ": "_",
            "#": "sh",
            ":": "_",
            "α": "a",
            "β": "b",
            "γ": "g",
            " ": "",
            "+": "pl",
            "/": "_",
            ":": "_",
            "-": "_",
            ".": "_",
            "?": "unkn",
            ",": "_",
            # "(": "__",
            # ")": "__",
            # "[": "__",
            # "]": "__",
            "(": "",
            ")": "",
            "[": "",
            "]": "",
            ">": "_",
            "<": "_",
        }

        for element in sbml2BnglTranslationDict:
            name = name.replace(element, sbml2BnglTranslationDict[element])
        name = re.sub("[\W]", "", name)
        return name

    def parseAnnotation(self, annotation):
        speciesAnnotationDict = defaultdict(list)
        lista = libsbml.CVTermList()
        libsbml.RDFAnnotationParser.parseRDFAnnotation(annotation, lista)
        # print '----',species.getName()
        for idx in range(0, lista.getSize()):
            for idx2 in range(0, lista.get(idx).getResources().getLength()):
                resource = lista.get(idx).getResources().getValue(idx2)
                qualifierType = lista.get(idx).getQualifierType()
                qualifierDescription = (
                    bioqual[lista.get(idx).getBiologicalQualifierType()]
                    if qualifierType
                    else modqual[lista.get(idx).getModelQualifierType()]
                )
                speciesAnnotationDict[qualifierDescription].append(resource)
        return speciesAnnotationDict

    def buildAnnotationDict(self, document):
        annotationDict = defaultdict(lambda: defaultdict(list))
        speciesNameDict = {}
        for species in document.getModel().getListOfSpecies():
            annotation = species.getAnnotation()
            transformedName = self.standardizeName(species.getName())
            speciesNameDict[transformedName] = species.getName()
            speciesNameDict[species.getName()] = transformedName
            if annotation:
                annotationDict[transformedName] = self.parseAnnotation(annotation)
        return annotationDict, speciesNameDict

    def updateFromParent(self, child, parent, annotationDict):
        for annotationLabel in annotationDict[parent]:
            if annotationLabel in ["BQB_IS_VERSION_OF", "BQB_IS"]:
                annotationDict[child]["BQB_IS_VERSION_OF"] = annotationDict[parent][
                    annotationLabel
                ]

    def updateFromChild(self, parent, child, annotationDict):
        for annotationLabel in annotationDict[child]:
            if annotationLabel in ["BQB_IS_VERSION_OF", "BQB_IS"]:
                annotationDict[parent]["BQB_HAS_VERSION"] = annotationDict[child][
                    annotationLabel
                ]

    def updateFromComplex(
        self, complexMolecule, sct, annotationDict, annotationToSpeciesDict
    ):
        localSpeciesDict = {}
        unmatchedReactants = []
        unmatchedAnnotations = []
        for constituentElement in sct[complexMolecule][0]:
            flag = False
            if len(annotationDict[constituentElement]) > 0:
                for annotation in annotationDict[constituentElement]:
                    if annotation in ["BQB_IS_VERSION_OF", "BQB_IS", "BQB_HAS_VERSION"]:
                        flag = True
                        for individualAnnotation in annotationDict[constituentElement][
                            annotation
                        ]:
                            localSpeciesDict[individualAnnotation] = constituentElement
                            localSpeciesDict[constituentElement] = individualAnnotation
                if flag:
                    continue

            if constituentElement in annotationToSpeciesDict:
                localSpeciesDict[constituentElement] = annotationToSpeciesDict[
                    constituentElement
                ]
                localSpeciesDict[
                    annotationToSpeciesDict[constituentElement]
                ] = constituentElement
            else:
                unmatchedReactants.append(constituentElement)

        for annotationType in annotationDict[complexMolecule]:
            if annotationType in ["BQB_HAS_VERSION", "BQB_HAS_PART"]:
                for constituentAnnotation in annotationDict[complexMolecule][
                    annotationType
                ]:
                    if constituentAnnotation not in localSpeciesDict:
                        unmatchedAnnotations.append(constituentAnnotation)
        if len(set(unmatchedReactants)) == 1 and len(set(unmatchedAnnotations)) == 1:
            localSpeciesDict[unmatchedReactants[0]] = unmatchedAnnotations[0]
            localSpeciesDict[unmatchedAnnotations[0]] = unmatchedReactants[0]
            annotationDict[unmatchedReactants[0]]["BQB_IS_VERSION_OF"] = [
                unmatchedAnnotations[0]
            ]

        elif len(unmatchedReactants) > 0 or len(unmatchedAnnotations) > 0:
            # annotate from database names
            # print '**//',complexMolecule,unmatchedReactants,unmatchedAnnotations
            pass

        for element in localSpeciesDict:
            if element not in annotationToSpeciesDict:
                annotationToSpeciesDict[element] = localSpeciesDict[element]

    def updateFromComponents(
        self, complexMolecule, sct, annotationDict, annotationToSpeciesDict
    ):
        localSpeciesDict = {}
        unmatchedReactants = []
        for constituentElement in sct[complexMolecule][0]:
            flag = False
            if len(annotationDict[constituentElement]) > 0:
                for annotation in annotationDict[constituentElement]:
                    if annotation in [
                        "BQB_IS_VERSION_OF",
                        "BQB_IS",
                        "BQB_HAS_VERSION",
                        "BQB_HAS_PART",
                    ]:
                        for individualAnnotation in annotationDict[constituentElement][
                            annotation
                        ]:
                            # localSpeciesDict[individualAnnotation] = constituentElement
                            localSpeciesDict[constituentElement] = individualAnnotation
                            flag = True
            if not flag:
                unmatchedReactants.append(constituentElement)
        for element in localSpeciesDict:
            annotationDict[complexMolecule]["BQB_HAS_PART"].append(
                localSpeciesDict[element]
            )

    def buildAnnotationTree(self, annotationDict, sct, database):
        annotationToSpeciesDict = {}
        for element in database.weights:
            if len(sct[element[0]]) > 0:

                if len(sct[element[0]][0]) == 1:
                    buildingBlock = sct[element[0]][0][0]
                    if len(annotationDict[element[0]]) == 0:
                        if len(annotationDict[buildingBlock]) > 0:
                            self.updateFromParent(
                                element[0], buildingBlock, annotationDict
                            )
                    if len(annotationDict[buildingBlock]) == 0:
                        if len(annotationDict[element[0]]) > 0:
                            self.updateFromChild(
                                buildingBlock, element[0], annotationDict
                            )
                elif len(sct[element[0]][0]) > 1:
                    if len(annotationDict[element[0]]) == 0:
                        self.updateFromComponents(
                            element[0], sct, annotationDict, annotationToSpeciesDict
                        )
                    else:
                        if (
                            "BQB_HAS_VERSION" in annotationDict[element[0]]
                            or "BQB_HAS_PART" in annotationDict[element[0]]
                        ):
                            self.updateFromComplex(
                                element[0], sct, annotationDict, annotationToSpeciesDict
                            )
                        else:
                            self.updateFromComponents(
                                element[0], sct, annotationDict, annotationToSpeciesDict
                            )
                #        annotationdict[element[0]]
                # for buildingBlock in sct[element[0]][0]:
                #    print '\t',buildingBlock,annotationDict[buildingBlock]

    def obtainSCT(
        self,
        fileName,
        reactionDefinitions,
        useID,
        namingConventions,
        speciesEquivalences=None,
    ):
        """
        one of the library's main entry methods. Process data from a file
        to obtain the species composition table, a dictionary describing
        the chemical history of different elements in the system
        """
        reader = libsbml.SBMLReader()
        document = reader.readSBMLFromFile(fileName)

        parser = SBML2BNGL(document.getModel(), useID)
        database = structures.Databases()
        database.forceModificationFlag = True
        database = mc.createSpeciesCompositionGraph(
            parser,
            database,
            reactionDefinitions,
            namingConventions,
            speciesEquivalences=speciesEquivalences,
            bioGridFlag=False,
        )

        return database.prunnedDependencyGraph, database, document

    def getAnnotationSystem(self):

        annotationDict, speciesNameDict = self.buildAnnotationDict(self.sbmlDocument)
        # self.buildAnnotationTree(annotationDict, self.sct, self.database)

        return annotationDict

    def getModelAnnotations(self):

        model = self.sbmlDocument.getModel()
        annotationXML = model.getAnnotation()
        lista = libsbml.CVTermList()
        libsbml.RDFAnnotationParser.parseRDFAnnotation(annotationXML, lista)
        modelAnnotations = []
        for idx in range(lista.getSize()):
            for idx2 in range(lista.get(idx).getResources().getLength()):
                if lista.get(idx).getQualifierType():
                    modelAnnotations.append(
                        [
                            bioqual[lista.get(idx).getBiologicalQualifierType()],
                            lista.get(idx).getResources().getValue(idx2),
                        ]
                    )
                else:
                    modelAnnotations.append(
                        [
                            modqual[lista.get(idx).getModelQualifierType()],
                            lista.get(idx).getResources().getValue(idx2),
                        ]
                    )

        return modelAnnotations


def defineConsole():
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-i", "--input-file", type=str, help="input SBML file", required=True
    )
    # parser.add_argument('-o','--output-file',type=str,help='output SBML file',required=True)
    return parser


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()
    # input_file = '/home/proto/workspace/bionetgen/parsers/SBMLparser/XMLExamples/curated/BIOMD%010i.xml' % 19
    annotationExtractor = AnnotationExtractor(namespace.input_file)
    modelAnnotations = annotationExtractor.getModelAnnotations()
    # elementalMolecules = [x for x in annotationExtractor.sct if annotationExtractor.sct[x] == []]
    complexMolecules = [
        x for x in annotationExtractor.sct if annotationExtractor.sct[x] != []
    ]
    print({x: annotationExtractor.getAnnotationSystem()[x] for x in complexMolecules})



from bionetgen.core.utils.logging import BNGLogger


class BNGInfo:
    """
    Used by the Cement app to execute the info subcommand, which involves gathering, preparing, and printing
    relevant version and path information.

    The gatherInfo() method finds and stores all necessary information.
    The messageGeneration() method prepares this information in a text string.
    The run() method simply outputs all the information.

    Requires a configuration file. An additional set of arguments are optional.
    """

    def __init__(self, config, args=None, app=None):
        self.config = config
        self.args = args
        self.app = app
        self.logger = BNGLogger(app=self.app)

    def gatherInfo(self):
        """
        Gathers information about relevant versions and paths into a dictionary.
        """
        import subprocess
        import os
        import bionetgen
        import numpy
        import pandas
        import roadrunner

        self.logger.debug("Gathering info", loc=f"{__file__} : BNGInfo.gatherInfo()")

        self.info = {}

        # Add some description for the following information
        self.info["\nThe following are related to BioNetGen and its execution"] = ""

        self.logger.debug("BNG info", loc=f"{__file__} : BNGInfo.gatherInfo()")
        # Get BNG version
        with open(
            os.path.join(
                *[os.path.dirname(bionetgen.__file__), "assets", "BNGVERSION"]
            ),
            "r",
        ) as f:
            read_data = f.read()
        self.info["BNG version"] = read_data[10:15]

        # Get BNG2.pl path
        self.info["BNG2.pl path"] = (
            self.config.get("bionetgen", "bngpath") + " (the main executable for BNG)"
        )

        self.logger.debug("Perl info", loc=f"{__file__} : BNGInfo.gatherInfo()")
        # Get Perl version
        # Read in CLI text
        result = subprocess.run(["perl", "-v"], stdout=subprocess.PIPE)
        text = str(result.stdout)
        # Find start & end indices
        num_start = text.find("(v") + 2
        num_end = text.find(")")
        # Save version info
        self.info["Perl version"] = text[num_start:num_end] + " (used to run BNG2.pl)"

        self.logger.debug("PyBNG info", loc=f"{__file__} : BNGInfo.gatherInfo()")
        # Get CLI version
        with open(
            os.path.join(*[os.path.dirname(bionetgen.__file__), "assets", "VERSION"]),
            "r",
        ) as f:
            read_data = f.read()
        self.info["CLI version"] = read_data

        # Get pyBNG path
        self.info["pyBNG path"] = (
            os.path.dirname(bionetgen.__file__) + " (the PyBNG installation)"
        )

        self.logger.debug(
            "Info on installed python libraries",
            loc=f"{__file__} : BNGInfo.gatherInfo()",
        )

        # Add some description for the following information
        self.info["\nThe following libraries are required by PyBioNetGen"] = ""

        # Get numpy version
        self.info["numpy version"] = numpy.version.version

        # Get pandas version
        self.info["pandas version"] = pandas.__version__

        # Get libRoadRunner version
        text = roadrunner.getVersionStr()
        self.info["libRoadRunner version"] = text[0:5]

        return self.info

    def messageGeneration(self):
        """
        Takes the dictionary created by gatherInfo() and
        converts it to a string of text for printing.
        """
        self.logger.debug(
            "Generating message", loc=f"{__file__} : BNGInfo.messageGeneration()"
        )

        self.message = " "

        # Create lines of message using self.info dictionary
        message_lines = []

        for key in self.info:
            text = str(key + ": " + self.info[key] + "\n")
            message_lines.append(text)
        # Join lines to create entire message
        self.message = "".join((message_lines))

        return self.message

    def run(self):
        """
        Simply prints out the created information message.
        """
        self.logger.debug("Printing message", loc=f"{__file__} : BNGInfo.run()")

        print(self.message)



import subprocess, os
from bionetgen.core.tools import BNGInfo
from bionetgen.core.tools import BNGVisualize
from bionetgen.core.tools import BNGCLI
from bionetgen.core.tools import BNGGdiff
from bionetgen.core.notebook import BNGNotebook
from bionetgen.core.utils.utils import run_command

import os, sys


# TODO Consolidate how config is being accessed. It's
# almost like each function accesses the configs from
# a different path
def runCLI(app):
    """
    Convenience function to run BNG2.pl from the CLI app

    Usage: runCLI(config, args)

    Arguments
    ---------
    config : dict
        configuration dictionary from BioNetGen cement app
    args :  argparse.Namespace
        arguments parsed from the command line with argparser.
    """
    args = app.pargs
    config = app.config
    # this pulls out the arguments
    sys.tracebacklimit = args.traceback_depth
    inp_file = args.input
    output = args.output
    log_file = args.log_file
    # if you set args.bngpath it should take precedence
    app.log.debug("Pulling BNG path from config", f"{__file__} : runCLI()")
    config_bngpath = config.get("bionetgen", "bngpath")
    # and instantiates the CLI object
    app.log.debug("Instantiating BNGCLI object", f"{__file__} : runCLI()")
    cli = BNGCLI(inp_file, output, config_bngpath, log_file=log_file, app=app)
    cli.stdout = config.get("bionetgen", "stdout")
    cli.stderr = config.get("bionetgen", "stderr")
    app.log.debug("Running", f"{__file__} : runCLI()")
    cli.run()


def plotDAT(app):
    """
    Convenience function to plot dat/scan files from the CLI

    Usage: plotDAT(inp, out, kw)

    Arguments
    ---------
    inp : str
        input gdat/cdat/scan file to plot
    out : str
        (optional) output file path, can be used to define the
        output format as well. Default is the current folder,
        filename is the same as the input file and default format
        is PNG.
    kw : dict
        (optional) this is a set of keyword arguments you want to
        pass for certain matplotlib options. Check -h for details
    """
    args = app.pargs
    # we need to have gdat/cdat files
    # TODO: Transition to BNGErrors and logging
    assert (
        args.input.endswith(".gdat")
        or args.input.endswith(".cdat")
        or args.input.endswith(".scan")
    ), "Input file has to be either a gdat or a cdat file"
    inp = args.input
    out = args.output
    kw = dict(args._get_kwargs())
    # if we want to plot directly into the folder
    # we are in we need to get the path correctly
    if out == ".":
        path, fname = os.path.split(inp)
        fnoext, ext = os.path.splitext(fname)
        out = os.path.join(path, "{}.png".format(fnoext))
    # use the plotter object to get the plot
    from bionetgen.core.tools import BNGPlotter

    app.log.debug("Instantiating BNGPlotter object", f"{__file__} : plotDAT()")
    plotter = BNGPlotter(inp, out, app=app, **kw)
    app.log.debug("Plotting", f"{__file__} : plotDAT()")
    plotter.plot()


def runAtomizeTool(app):
    """
    Uses AtomizeTool class to run atomizer from a set of arguments
    """
    # pull args/config
    args = app.pargs
    config = app.config
    # run AtomizeTool
    from bionetgen.atomizer import AtomizeTool

    app.log.debug("Instantiating AtomizeTool object", f"{__file__} : runAtomizeTool()")
    a = AtomizeTool(parser_namespace=args, app=app)
    # do config specific stuff here if need be, or remove the config requirement
    app.log.debug("Atomizing", f"{__file__} : runAtomizeTool()")
    a.run()


def printInfo(app):
    """
    Uses BNGInfo class to print BioNetGen information using
    arguments and config from Cement framework.
    """
    config = app.config
    app.log.debug("Instantiating BNGInfo object", f"{__file__} : printInfo()")
    info = BNGInfo(config=config, app=app)
    app.log.debug("Gathering and printing info", f"{__file__} : printInfo()")
    info.gatherInfo()
    info.messageGeneration()
    info.run()


def visualizeModel(app):
    """
    Uses BNGVisualize class to visualize BNGL models using
    arguments and configuration from Cement framework.
    """
    # pull args/config from app
    args = app.pargs
    config = app.config
    # pull relevant arguments for the tool
    inp = args.input
    out = args.output
    vtype = args.type
    app.log.debug("Pulling BNG path from config", f"{__file__} : visualizeModel()")
    # if you set args.bngpath it should take precedence
    config_bngpath = config.get("bionetgen", "bngpath")
    # run visualize tool
    app.log.debug("Instantiating BNGVisualize object", f"{__file__} : visualizeModel()")
    viz = BNGVisualize(inp, output=out, vtype=vtype, bngpath=config_bngpath, app=app)
    app.log.debug("Visualizing", f"{__file__} : visualizeModel()")
    viz.run()


def graphDiff(app):
    """
    Uses BNGGdiff object to calculate differences between two graphs
    (only works with graphml files generated by BioNetGen)
    """
    # pull args and config for the tool
    args = app.pargs
    # if you set args.bngpath it should take precedence
    app.log.debug("Instantiating BNGGdiff object", f"{__file__} : graphDiff()")
    gdiff = BNGGdiff(
        args.input,
        args.input2,
        out=args.output,
        out2=args.output2,
        mode=args.mode,
        colors=args.colors,
        app=app,
    )
    app.log.debug("Calculating graph diff", f"{__file__} : graphDiff()")
    gdiff.run()


def generate_notebook(app):
    """
    Uses BNGNotebook class to write a Jupyter notebook from a
    given set of command line arguments
    """
    args = app.pargs
    if args.input is not None:
        # we want to use the template to write a custom notebok
        # TODO: Transition to BNGErrors and logging
        assert args.input.endswith(
            ".bngl"
        ), f"File {args.input} doesn't have bngl extension!"
        try:
            app.log.debug("Loading model", f"{__file__} : notebook()")
            import bionetgen

            m = bionetgen.bngmodel(args.input)
            str(m)
        except:
            app.log.error("Failed to load model", f"{__file__} : notebook()")
            raise RuntimeError(f"Couldn't import given model: {args.input}!")
        notebook = BNGNotebook(
            app.config["bionetgen"]["notebook"]["template"],
            INPUT_ARG=args.input,
        )
    else:
        # just use the basic notebook
        notebook = BNGNotebook(app.config["bionetgen"]["notebook"]["path"])
    # find our file name
    if len(args.output) == 0:
        fname = app.config["bionetgen"]["notebook"]["name"]
    else:
        fname = args.output
    # write the notebook out
    if os.path.isdir(fname):
        if args.input is not None:
            basename = os.path.basename(args.input)
            mname = basename.replace(".bngl", "")
            fname = mname + ".ipynb"
        else:
            mname = app.config["bionetgen"]["notebook"]["name"]
            fname = os.path.join(args.output, mname)

    app.log.debug(f"Writing notebook to file: {fname}", f"{__file__} : notebook()")
    notebook.write(fname)
    # open the notebook with nbopen
    # TODO: deal with stdout/err
    app.log.debug(
        f"Attempting to open notebook {fname} with nbopen",
        f"{__file__} : notebook()",
    )
    stdout = getattr(subprocess, app.config["bionetgen"]["stdout"])
    stderr = getattr(subprocess, app.config["bionetgen"]["stderr"])
    if args.open:
        command = ["nbopen", fname]
        rc, _ = run_command(command)



# -*- coding: utf-8 -*-
"""
Created on Thu Nov 28 17:17:12 2013

@author: proto
"""
# import sys
# sys.path.insert(0, '../utils/')
import utils.consoleCommands as console
from .utils import readBNGXML
import networkx as nx
import cPickle as pickle
from collections import Counter
from os import listdir


def simpleGraph(graph, species, observableList, prefix="", superNode={}):
    """
    build a simple contact map. species, components and addbond
    information is included
    """
    edgeList = []
    nodeDict = {}
    for singleSpecies in species:
        if "{1}_{0}".format(singleSpecies.name, prefix) in superNode:
            name = superNode["{1}_{0}".format(singleSpecies.name, prefix)]
            graph.add_node(name, size=superNode[name], color="orange")
        else:
            name = "{1}_{0}".format(singleSpecies.name, prefix)
            graph.add_node(name, size=1, color="pink")

        nodeDict[singleSpecies.idx] = name
        for component in singleSpecies.components:
            graph.add_node(
                "{0}({1})".format(name, component.name), size=0.5, color="teal"
            )
            edgeList.append(
                ["{0}".format(name), "{0}({1})".format(name, component.name), 10]
            )
    for edge in observableList:
        if len(edge) == 2:
            name1 = "{0}_{1}".format(prefix, edge[0].split("(")[0])
            name2 = "{0}_{1}".format(prefix, edge[1].split("(")[0])
            if name1 in superNode:
                name1 = superNode[name1]
            if name2 in superNode:
                name2 = superNode[name2]
            name1 = "{0}({1}".format(name1, edge[0].split("(")[1])
            name2 = "{0}({1}".format(name2, edge[1].split("(")[1])

            edgeList.append([name1, name2, 2])
    graph.add_weighted_edges_from(edgeList)
    return nodeDict


def main():
    with open("linkArray.dump", "rb") as f:
        linkArray = pickle.load(f)
    with open("xmlAnnotationsExtended.dump", "rb") as f:
        annotations = pickle.load(f)

    speciesEquivalence = {}
    onlyDicts = [x for x in listdir("./complex")]
    onlyDicts = [x for x in onlyDicts if ".bngl.dict" in x]

    for x in onlyDicts:
        with open("complex/{0}".format(x), "rb") as f:
            speciesEquivalence[int(x.split(".")[0][6:])] = pickle.load(f)

    for cidx, cluster in enumerate(linkArray):
        # FIXME:only do the first cluster
        cidx = 0
        cluster = linkArray[0]
        if len(cluster) == 1:
            continue
        annotationsDict = {idx: x for idx, x in enumerate(annotations)}
        annotationsIncidence = {}

        # get the model intersection points

        for element in annotationsDict:
            for species in annotationsDict[element]:
                key = annotationsDict[element][species][1]
                if key not in annotationsIncidence:
                    annotationsIncidence[key] = set([])
                annotationsIncidence[key].add(element)

        # number = 19
        graph = nx.Graph()
        nodeDict = {}
        superNodeDict = {}

        for element in annotationsIncidence:
            if len(annotationsIncidence[element]) > 1:
                tmpEdgeList = []
                names = []
                for idx in annotationsIncidence[element]:
                    for species in [
                        x
                        for x in annotationsDict[idx]
                        if annotationsDict[idx][x][1] == element
                    ]:
                        try:
                            print(speciesEquivalence[idx])
                            if isinstance(speciesEquivalence[idx][species], str):
                                tmpEdgeList.append(
                                    "{0}_{1}".format(
                                        idx, speciesEquivalence[idx][species][0:-2]
                                    )
                                )
                                names.append(speciesEquivalence[idx][species][0:-2])
                            else:
                                tmpEdgeList.append(
                                    "{0}_{1}".format(
                                        idx,
                                        speciesEquivalence[idx][species]
                                        .molecules[0]
                                        .name,
                                    )
                                )
                                names.append(
                                    speciesEquivalence[idx][species].molecules[0].name
                                )
                        except:
                            print("----", "error", idx, species)
                            pass
                names = [x.lower() for x in names]
                name = Counter(names).most_common(5)
                name = [x for x in name if len(x[0]) > 1]
                if len(name) == 0:
                    name = Counter(names).most_common(1)
                for edge in tmpEdgeList:
                    superNodeDict[edge] = name[0][0]
                if len(name) > 0:
                    superNodeDict[name[0][0]] = len(tmpEdgeList)
                # pairs = [(x, y, 1) for x in tmpEdgeList for y in tmpEdgeList if x!= y]
                # graph.add_weighted_edges_from(pairs)
        print(superNodeDict)
        for x in cluster:
            number = x - 1
            try:
                console.bngl2xml("complex/output{0}.bngl".format(number))
                species, rules, parameterDict, observableList = readBNGXML.parseXML(
                    "output{0}.xml".format(number)
                )

            except:
                continue
            nodeDict[number] = simpleGraph(
                graph, species, observableList, number, superNodeDict
            )

        nx.write_gml(graph, "cgraph_{0}.gml".format(cidx))
        break


def main2():
    number = 48
    species, rules, parameterDict, observableList = readBNGXML.parseXML(
        "output{0}.xml".format(number)
    )
    graph = nx.Graph()

    simpleGraph(graph, species, observableList, number)
    nx.write_gml(graph, "graph_{0}.gml".format(number))


if __name__ == "__main__":
    main()



from bionetgen.main import BioNetGen
from bionetgen.network.networkparser import BNGNetworkParser
from bionetgen.network.blocks import (
    NetworkGroupBlock,
    NetworkParameterBlock,
    NetworkReactionBlock,
    NetworkSpeciesBlock,
    NetworkCompartmentBlock,
    NetworkFunctionBlock,
    NetworkEnergyPatternBlock,
    NetworkPopulationMapBlock,
)


# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]

###### CORE OBJECT AND PARSING FRONT-END ######
class Network:
    """
    Main model object and entry point for model API. The goal of this
    object is to generate and read the BNGXML of a given BNGL model
    and give the user a pythonic interface to the resulting model object.

    Usage: bngmodel(bng_model)
           bngmodel(bng_model, BNGPATH)

    Attributes
    ----------
    active_blocks : list[str]
        a list of the blocks that have been parsed in the model
    bngnetworkparser : BNGNetworkParser
        BNGParser object that's responsible for .bngl file reading and model setup
    network_name : str
        name of the model, generally set from the given BNGL file

    Methods
    -------
    write_model(model_name)
        write the model in BNGL format to the path given
    setup_simulator(sim_type)
        sets up a simulator in bngmodel.simulator where the only current supported
        type of simulator is libRR for libRoadRunner simulator.
    """

    def __init__(self, bngl_model, BNGPATH=def_bng_path):
        self.active_blocks = []
        # We want blocks to be printed in the same order every time
        self.block_order = [
            "parameters",
            "species",
            "reactions",
            "groups"
            # "compartments",
            # "molecule_types",
            # "species",
            # "functions",
            # "energy_patterns",
            # "population_maps",
            # "actions",
        ]
        self.network_name = ""
        self.bngnetworkparser = BNGNetworkParser(bngl_model)
        self.bngnetworkparser.parse_network(self)
        for block in self.block_order:
            if block not in self.active_blocks:
                self.add_empty_block(block)
        # Check to see if there are no active blocks
        # If not, model is most likely not in BNGL format
        if not self.active_blocks:
            print(
                "WARNING: No active blocks. Please ensure model is in proper BNGL or BNG-XML format"
            )

    def __str__(self):
        """
        write the model to str
        """
        model_str = ""
        for block in self.block_order:
            # ensure we didn't get new items into a
            # previously inactive block, if we did
            # add them to the active blocks
            if hasattr(self, block):
                if len(getattr(self, block)) > 0:
                    if getattr(self, block).name not in self.active_blocks:
                        self.active_blocks.append(block)
                # if we removed items from a block and
                # it's now empty, we want to remove it
                # from the active blocks
                elif len(getattr(self, block)) == 0 and block in self.active_blocks:
                    self.active_blocks.remove(block)
            # print only the active blocks
            if block in self.active_blocks:
                if block != "actions" and len(getattr(self, block)) > 0:
                    model_str += str(getattr(self, block))
        return model_str

    def __repr__(self):
        return self.network_name

    def __iter__(self):
        active_ordered_blocks = [
            getattr(self, i) for i in self.block_order if i in self.active_blocks
        ]
        return active_ordered_blocks.__iter__()

    def add_block(self, block):
        bname = block.name.replace(" ", "_")
        # TODO: fix this exception
        block_adder = getattr(self, "add_{}_block".format(bname))
        block_adder(block)

    def add_empty_block(self, block_name):
        bname = block_name.replace(" ", "_")
        # TODO: fix this exception
        block_adder = getattr(self, "add_{}_block".format(bname))
        block_adder()

    def add_parameters_block(self, block=None):
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, NetworkParameterBlock)
            self.parameters = block
            if "parameters" not in self.active_blocks:
                self.active_blocks.append("parameters")
        else:
            self.parameters = NetworkParameterBlock()

    # def add_compartments_block(self, block=None):
    #     if block is not None:
    #         assert isinstance(block, NetworkCompartmentBlock)
    #         self.compartments = block
    #         if "compartments" not in self.active_blocks:
    #             self.active_blocks.append("compartments")
    #     else:
    #         self.compartments = NetworkCompartmentBlock()

    def add_species_block(self, block=None):
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, NetworkSpeciesBlock)
            self.species = block
            if "species" not in self.active_blocks:
                self.active_blocks.append("species")
        else:
            self.species = NetworkSpeciesBlock()

    def add_groups_block(self, block=None):
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, NetworkGroupBlock)
            self.groups = block
            if "groups" not in self.active_blocks:
                self.active_blocks.append("groups")
        else:
            self.groups = NetworkGroupBlock()

    def add_reactions_block(self, block=None):
        if block is not None:
            # TODO: Transition to BNGErrors and logging
            assert isinstance(block, NetworkReactionBlock)
            self.reactions = block
            if "reactions" not in self.active_blocks:
                self.active_blocks.append("reactions")
        else:
            self.reactions = NetworkReactionBlock()

    # def add_functions_block(self, block=None):
    #     if block is not None:
    #         assert isinstance(block, NetworkFunctionBlock)
    #         self.functions = block
    #         if "functions" not in self.active_blocks:
    #             self.active_blocks.append("functions")
    #     else:
    #         self.functions = NetworkFunctionBlock()

    # def add_energy_patterns_block(self, block=None):
    #     if block is not None:
    #         assert isinstance(block, NetworkEnergyPatternBlock)
    #         self.energy_patterns = block
    #         if "energy_patterns" not in self.active_blocks:
    #             self.active_blocks.append("energy_patterns")
    #     else:
    #         self.energy_patterns = NetworkEnergyPatternBlock()

    # def add_population_maps_block(self, block=None):
    #     if block is not None:
    #         assert isinstance(block, NetworkPopulationMapBlock)
    #         self.population_maps = block
    #         if "population_maps" not in self.active_blocks:
    #             self.active_blocks.append("population_maps")
    #     else:
    #         self.population_maps = NetworkPopulationMapBlock()

    def write_model(self, file_name):
        """
        write the model to file
        """
        model_str = ""
        for block in self.active_blocks:
            model_str += str(getattr(self, block))
        with open(file_name, "w") as f:
            f.write(model_str)



class RuleMod:
    """
    Rule modifiers class for storage and printing.
    """

    def __init__(self, mod_type=None) -> None:
        # valid mod types
        self.valid_mod_names = ["DeleteMolecules", "MoveConnected", "TotalRate"]
        self.type = mod_type

    def __str__(self) -> str:
        if self.type is None:
            return ""
        else:
            return self.type

    def __repr__(self) -> str:
        return f"Rule modifier of tyoe {self.type}"

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, val):
        if val in self.valid_mod_names or val is None:
            self._type = val
        else:
            print(f"Rule modifier type {val} is not a valid type")



# -*- coding: utf-8 -*-
"""
Created on Sun Mar 25 21:26:49 2012

@author: proto
"""
from __future__ import division
import json
from functools import partial
import marshal, colorlog
from bionetgen.core.utils.logging import BNGLogger

from pyparsing import (
    Literal,
    CaselessLiteral,
    Word,
    Combine,
    Group,
    Optional,
    ZeroOrMore,
    Forward,
    nums,
    alphas,
    oneOf,
    alphanums,
)
import math
import operator
import os
from subprocess import call
import functools
import pylru

logger = BNGLogger()


def pmemoize(obj):
    cache = obj.cache = pylru.lrucache(500)

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = marshal.dumps([args, kwargs])
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]

    return memoizer


# import progressbar

# sys.path.insert(0, '../utils/')
# import consoleCommands as console


class memoize(object):
    """cache the return value of a method

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method:
    class Obj(object):
        @memoize
        def add_to(self, arg):
            return self + arg
    Obj.add_to(1) # not enough arguments
    Obj.add_to(1, 2) # returns 3, result is not cached
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, marshal.dumps(args[1:]), frozenset(kw.items()))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


# ASS: optimizing the memoization method used with
# resolveHelper method specifically to use less
# memory, important in certain larger models
class memoizeMapped(object):
    """
    Optimized local cache for recursive resolveHelper method
    to limit memory usage
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, obj, gkey, react, mem, withMod):
        key1 = gkey
        key2 = react
        # This memory hash is a bit suspect,
        key3 = hash(tuple(sorted(mem)))
        key4 = withMod
        try:
            res = self.cache[key1][key2][key3][key4]
        except KeyError:
            try:
                d = self.cache[key1]
            except KeyError:
                self.cache[key1] = {}
            try:
                d = self.cache[key1][key2]
            except KeyError:
                self.cache[key1][key2] = {}
            try:
                d = self.cache[key1][key2][key3]
            except KeyError:
                self.cache[key1][key2][key3] = {}
            try:
                d = self.cache[key1][key2][key3][key4]
            except KeyError:
                self.cache[key1][key2][key3][key4] = None
            res = self.cache[key1][key2][key3][key4] = self.func(
                obj, gkey, react, mem, withMod
            )
        return res


class TranslationException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class NumericStringParser(object):

    """
    Most of this code comes from the fourFn.py pyparsing example

    """

    def pushFirst(self, strg, loc, toks):
        self.exprStack.append(toks[0])

    def pushUMinus(self, strg, loc, toks):
        if toks and toks[0] == "-":
            self.exprStack.append("unary -")

    def __init__(self):
        """
        expop   :: '^'
        multop  :: '*' | '/'
        addop   :: '+' | '-'
        integer :: ['+' | '-'] '0'..'9'+
        atom    :: PI | E | real | fn '(' expr ')' | '(' expr ')'
        factor  :: atom [ expop factor ]*
        term    :: factor [ multop factor ]*
        expr    :: term [ addop term ]*
        """
        point = Literal(".")
        e = CaselessLiteral("E")
        fnumber = Combine(
            Word("+-" + alphanums + "_", alphanums + "_")
            + Optional(point + Optional(Word(alphanums + "_")))
            + Optional(e + Word("+-" + alphanums + "_", alphanums + "_"))
        )

        ident = Word(alphas, alphanums + "_")
        plus = Literal("+")
        minus = Literal("-")
        mult = Literal("*")
        div = Literal("/")
        lpar = Literal("(").suppress()
        rpar = Literal(")").suppress()

        addop = plus | minus
        multop = mult | div
        expop = Literal("^")
        pi = CaselessLiteral("PI")
        expr = Forward()
        function = ident + lpar + expr + ZeroOrMore("," + expr) + rpar
        atom = (
            (
                Optional(oneOf("- +"))
                + (pi | e | function | fnumber).setParseAction(self.pushFirst)
            )
            | Optional(oneOf("- +")) + Group(lpar + expr + rpar)
        ).setParseAction(self.pushUMinus)
        # by defining exponentiation as "atom [ ^ factor ]..." instead of
        # "atom [ ^ atom ]...", we get right-to-left exponents, instead of left-to-right
        # that is, 2^3^2 = 2^(3^2), not (2^3)^2.
        factor = Forward()
        factor << atom + ZeroOrMore((expop + factor).setParseAction(self.pushFirst))
        term = factor + ZeroOrMore((multop + factor).setParseAction(self.pushFirst))
        expr << term + ZeroOrMore((addop + term).setParseAction(self.pushFirst))
        # addop_term = ( addop + term ).setParseAction( self.pushFirst )
        # general_term = term + ZeroOrMore( addop_term ) | OneOrMore( addop_term)
        # expr <<  general_term
        self.bnf = expr
        # map operator symbols to corresponding arithmetic operations
        epsilon = 1e-12
        self.opn = {
            "+": operator.add,
            "-": operator.sub,
            "*": operator.mul,
            "/": operator.truediv,
            "^": operator.pow,
        }
        self.fn = {
            "sin": math.sin,
            "cos": math.cos,
            "tan": math.tan,
            "abs": abs,
            "trunc": lambda a: int(a),
            "round": round,
            "sgn": lambda a: abs(a) > epsilon and cmp(a, 0) or 0,
        }

    def evaluateStack(self, s):
        op = s.pop()
        print(op)
        if op == "unary -":
            return -self.evaluateStack(s)
        if op in "+-*/^":
            op2 = self.evaluateStack(s)
            op1 = self.evaluateStack(s)
            return self.opn[op](op1, op2)
        elif op == "PI":
            return math.pi  # 3.1415926535
        elif op == "E":
            return math.e  # 2.718281828
        elif op in self.fn:
            return self.fn[op](self.evaluateStack(s))
        elif op[0].isalpha():
            return 0
        else:
            return float(op)

    def eval(self, num_string, parseAll=True):
        self.exprStack = []
        # remove comments
        if num_string.find("#") != -1:
            num_string = num_string[0 : num_string.find("#")]
        try:
            results = self.bnf.parseString(num_string, parseAll)
            return True
        except:
            return False

        # val=self.evaluateStack( self.exprStack[:] )
        # return val


def defaultReactionDefinition():
    listOfReactions = {"1": [["S0", "S1"], ["S2"]], "2": [["S2"], ["S0", "S1"]]}
    listOfDefinitions = {"Binding": [1, 2]}
    final = {"reactions": listOfReactions, "definitions": listOfDefinitions}
    with open("reactionDefinition.json", "w") as fp:
        json.dump(final, fp)


# def setupLog(fileName, level, quietMode=False):
#     if quietMode:
#         colorlog.basicConfig(filename=fileName, level=level, filemode="w")
#     else:
#         colorlog.basicConfig(level=level)


# def setupStreamLog(console):
#     # set colorlog handler
#     fmter = colorlog.ColoredFormatter(
#         "%(log_color)s%(levelname)s:%(name)s:%(message)s",
#         log_colors={
#             "DEBUG": "cyan",
#             "INFO": "green",
#             "WARNING": "yellow",
#             "ERROR": "red",
#             "CRITICAL": "red",
#         },
#     )
#     # tell the handler to use this format
#     console.setFormatter(fmter)
#     # colorlog.getLogger().addHandler(console)


# def finishStreamLog(console):
#     colorlog.getLogger().removeHandler(console)


def logMess(logType, logMessage):

    level = logType.split(":")[0]
    module = logType.split(":")[1]

    if level == "INFO":
        logger.info(logMessage, loc=f"{__file__} : {module}.logMess()")
    elif level == "DEBUG":
        logger.debug(logMessage, loc=f"{__file__} : {module}.logMess()")
    elif level == "WARNING":
        logger.warning(logMessage, loc=f"{__file__} : {module}.logMess()")
    elif level == "CRITICAL":
        logger.critical(logMessage, loc=f"{__file__} : {module}.logMess()")
    elif level == "ERROR":
        logger.error(logMessage, loc=f"{__file__} : {module}.logMess()")


def testBNGFailure(fileName):
    with open(os.devnull, "w") as f:
        result = call(["bngdev", fileName], stdout=f)
    return result


'''
def getValidFiles(directory, extension):
    """
    Gets a list of bngl files that could be correctly translated in a given 'directory'
    """
    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, '*.{0}'.format(extension)):
            matches.append(os.path.join(root, filename))
    for i in xrange(len(matches)):
        matches[i] = (matches[i], os.path.getsize(matches[i]))
    matches.sort(key=lambda filename: filename[1], reverse=False)
    matches = [x[0] for x in matches]
    return matches


def generateBNGXML(directory):

    bnglFiles = getValidFiles(directory, 'bngl')
    print('converting {0} bnglfiles'.format(len(bnglFiles)))
    progress = progressbar.ProgressBar()
    for i in progress(range(len(bnglFiles))):
        xmlName = '.'.join(bnglFiles[i].split('.')[:-1]) + '.xml'
        

        if os.path.exists(xmlName):
            continue
        console.bngl2xml(bnglFiles[i], timeout=10)

    print('moving xml files')
    files = glob.iglob(os.path.join('.', "*.xml"))
    for xmlfile in files:
        if os.path.isfile(xmlfile):
            shutil.move(xmlfile, directory)
'''



# -*- coding: utf-8 -*-
"""
Created on Mon Sep  2 18:11:35 2013

@author: proto
"""
import bionetgen


def setBngExecutable(executable):
    global bngExecutable
    bngExecutable = executable


def getBngExecutable():
    return bngExecutable


def bngl2xml(bnglFile, timeout=60):
    mdl = bionetgen.modelapi.bngmodel(bnglFile)
    xml_file = bnglFile.replace(".bngl", ".xml")
    with open(xml_file, "w+") as f:
        mdl.bngparser.bngfile.write_xml(f, xml_type="bngxml", bngl_str=str(mdl))
    # TODO: Deal with timeout here



import os, subprocess
from bionetgen.core.exc import BNGRunError
from bionetgen.core.utils.logging import BNGLogger


class BNGCLI:
    """
    Command Line Interface class to run BNG2.pl on a given
    model.

    Usage: BNGCLI(inp_file, output, bngpath)

    Arguments
    ---------
    inp_file : str
        path to the the BNGL file to run
    output : str
        path to the output folder to run the model in
    bngpath : str
        path to BioNetGen folder where BNG2.pl lives

    Methods
    -------
    run()
        runs the model in the given output folder
    """

    def __init__(
        self,
        inp_file,
        output,
        bngpath,
        suppress=False,
        log_file=None,
        timeout=None,
        app=None,
    ):
        self.app = app
        self.logger = BNGLogger(app=self.app)
        self.logger.debug(
            "Setting up BNGCLI object", loc=f"{__file__} : BNGCLI.__init__()"
        )
        self.inp_file = inp_file
        import bionetgen.modelapi.model as mdl

        if isinstance(inp_file, mdl.bngmodel):
            self.is_bngmodel = True
        else:
            self.is_bngmodel = False
            # ensure correct path to the input file
            self.inp_path = os.path.abspath(self.inp_file)
        # pull other arugments out
        self._set_output(output)
        # sedml_file = sedml
        self.bngpath = bngpath
        # setting up bng2.pl
        self.bng_exec = os.path.join(self.bngpath, "BNG2.pl")
        # TODO: Transition to BNGErrors and logging
        assert os.path.exists(self.bng_exec), "BNG2.pl is not found!"
        if "BNGPATH" in os.environ:
            self.old_bngpath = os.environ["BNGPATH"]
        else:
            self.old_bngpath = None
        os.environ["BNGPATH"] = self.bngpath
        self.result = None
        self.stdout = "PIPE"
        self.stderr = "STDOUT"
        self.suppress = suppress
        self.log_file = log_file
        self.timeout = timeout

    def _set_output(self, output):
        self.logger.debug(
            "Setting up output path", loc=f"{__file__} : BNGCLI._set_output()"
        )
        # setting up output area
        self.output = output
        if os.path.isdir(output):
            # path exists, let's go there
            os.chdir(output)
        else:
            os.mkdir(output)
            os.chdir(output)

    def run(self):
        self.logger.debug("Running", loc=f"{__file__} : BNGCLI.run()")
        from bionetgen.core.utils.utils import run_command

        try:
            stdout_loc = getattr(subprocess, self.stdout)
        except:
            stdout_loc = subprocess.PIPE
        try:
            stderr_loc = getattr(subprocess, self.stderr)
        except:
            stderr_loc = subprocess.STDOUT
        # run BNG2.pl
        if self.is_bngmodel:
            self.logger.debug(
                "The given model is a bngmodel object", loc=f"{__file__} : BNGCLI.run()"
            )
            self.logger.debug(
                "Writing the model to a file", loc=f"{__file__} : BNGCLI.run()"
            )
            write_to = self.inp_file.model_name + ".bngl"
            write_to = os.path.abspath(write_to)
            if os.path.isfile(write_to):
                self.logger.warning(
                    f"Overwriting file {write_to}", loc=f"{__file__} : BNGCLI.run()"
                )
            with open(write_to, "w") as tfile:
                tfile.write(str(self.inp_file))
            command = ["perl", self.bng_exec, write_to]
        else:
            self.logger.debug(
                "The given model is a file", loc=f"{__file__} : BNGCLI.run()"
            )
            fname = os.path.basename(self.inp_path)
            fname = fname.replace(".bngl", "")
            command = ["perl", self.bng_exec, self.inp_path]
        self.logger.debug("Running command", loc=f"{__file__} : BNGCLI.run()")
        rc, out = run_command(command, suppress=self.suppress, timeout=self.timeout)
        if self.log_file is not None:
            self.logger.debug("Setting up log file", loc=f"{__file__} : BNGCLI.run()")
            # test if we were given a path
            # TODO: This is a simple hack, might need to adjust it
            # trying to check if given file is an absolute/relative
            # path and if so, use that one. Otherwise, divine the
            # current path.
            if os.path.exists(self.log_file):
                # file or folder exists, check if folder
                if os.path.isdir(self.log_file):
                    fname = os.path.basename(self.inp_path)
                    fname = fname.replace(".bngl", "")
                    full_log_path = os.path.join(self.log_file, fname + ".log")
                else:
                    # it's intended to be file, so we keep it as is
                    full_log_path = self.log_file
            else:
                # doesn't exist, so we assume it's a file
                # and we keep it as is
                full_log_path = self.log_file
            self.logger.debug("Writing log file", loc=f"{__file__} : BNGCLI.run()")
            with open(full_log_path, "w") as f:
                f.write("\n".join(out))

        # if self.is_bngmodel:
        #     os.remove(tfile.name)
        # write out stdout/err if they exist
        # TODO Maybe indicate that we are printing out stdout/stderr before printing
        # if rc.stdout is not None:
        #     print(rc.stdout.decode('utf-8'))
        # if rc.stderr is not None:
        #     print(rc.stderr.decode('utf-8'))
        if rc == 0:
            self.logger.debug(
                "Command ran successfully", loc=f"{__file__} : BNGCLI.run()"
            )
            from bionetgen.core.tools import BNGResult

            # load in the result
            self.result = BNGResult(os.getcwd())
            self.result.process_return = rc
            self.result.output = out
            # set BNGPATH back
            if self.old_bngpath is not None:
                os.environ["BNGPATH"] = self.old_bngpath
        else:
            self.logger.error("Command failed to run", loc=f"{__file__} : BNGCLI.run()")
            self.result = None
            # set BNGPATH back
            if self.old_bngpath is not None:
                os.environ["BNGPATH"] = self.old_bngpath
            if hasattr(out, "stdout"):
                stdout_str = out.stdout.decode("utf-8")
            else:
                stdout_str = None
            if hasattr(out, "stdout"):
                stderr_str = out.stderr.decode("utf-8")
            else:
                stderr_str = None
            raise BNGRunError(command, stdout=stdout_str, stderr=stderr_str)



import argparse
import componentGroups
from utils import readBNGXML
import utils.structures as st
from collections import defaultdict
import pprint
import yaml

# moleculemapping = {'Epidermal_Growth_Factor': 'EGF'}
# componentmapping = {'Raf': {'mmod': 'astMod'}, 'JAK': {'mmod': 'genericMod'}}

moleculemapping = {}
componentmapping = {}

from collections import defaultdict


class prettyDict(defaultdict):
    __repr__ = dict.__repr__


def moleculeIntersection(nameSet1, nameSet2, mapping):
    intersection = {}
    for name1 in nameSet1:
        if name1 in nameSet2:
            intersection[name1] = name1
        elif name1 in mapping:
            intersection[name1] = mapping[name1]
    return intersection


class ModelComparison:
    def compareModelsStructurally(
        self, fileName1, fileName2, mapping, componentMapping
    ):
        molecules, _, _ = readBNGXML.parseXML(fileName1)
        molecules2, _, _ = readBNGXML.parseXML(fileName2)

        moleculeName1 = {x.name: x for x in molecules}
        moleculeName2 = {x.name: x for x in molecules2}
        intersectionDict = {}

        nameIntersection = moleculeIntersection(
            moleculeName1.keys(), moleculeName2.keys(), mapping
        )
        componentNameIntersection = defaultdict(lambda: defaultdict(str))
        for molecule in nameIntersection:

            for component in moleculeName1[molecule].components:
                if component.name in [
                    x.name for x in moleculeName2[nameIntersection[molecule]].components
                ]:
                    componentNameIntersection[nameIntersection[molecule]][
                        component.name
                    ] = component.name
                elif (
                    molecule in componentMapping
                    and component.name in componentMapping[molecule]
                ):
                    componentNameIntersection[nameIntersection[molecule]][
                        component.name.lower()
                    ] = componentMapping[molecule][component.name].lower()
        return nameIntersection, componentNameIntersection

    def discoverDifferences(self, contextMatrix, fileName1, fileName2):
        criticalComponents = prettyDict(list)
        totalCounter = 0.0
        similarityCounter = 0.0
        for molecule in contextMatrix:
            for component1 in contextMatrix[molecule]:
                for component2 in contextMatrix[molecule][component1]:
                    if (
                        len(contextMatrix[molecule][component1][component2]) != 2
                        or contextMatrix[molecule][component1][component2][fileName1]
                        != contextMatrix[molecule][component1][component2][fileName2]
                    ):
                        criticalComponents[molecule].append(
                            [
                                (component1, component2),
                                contextMatrix[molecule][component1][component2],
                            ]
                        )
                    else:
                        similarityCounter += 1.0
                    totalCounter += 1.0
        return criticalComponents, similarityCounter, totalCounter

    def hypothesisGenerator(self, premises, fileName1, fileName2):
        experiments = []
        serializedPremise = premises[1].items()
        if len(premises[1]) == 1:
            targetFileName = fileName2 if fileName1 in premises[1] else fileName1

            if serializedPremise[0][1] == "requirement":
                sourceMolecule = [premises[0][0], 1]
                targetMolecule = [premises[0][1], 0]
                experiments.append([targetFileName, sourceMolecule, targetMolecule])
            elif serializedPremise[0][1] == "nullrequirement":
                sourceMolecule = [premises[0][0], 1]
                targetMolecule = [premises[0][1], 1]
                experiments.append([targetFileName, sourceMolecule, targetMolecule])
            elif serializedPremise[0][1] == "exclusion":
                sourceMolecule = [premises[0][0], 1]
                targetMolecule = [premises[0][1], 1]
                experiments.append([targetFileName, sourceMolecule, targetMolecule])
                sourceMolecule = [premises[0][1], 1]
                targetMolecule = [premises[0][0], 1]
                experiments.append([targetFileName, sourceMolecule, targetMolecule])

        elif len(premises[1]) == 2:
            if serializedPremise[0][1] == "nullrequirement":
                if serializedPremise[1][1] == "exclusion":
                    targetFileName = serializedPremise[1][0]
                    sourceMolecule = [premises[0][1], 1]
                    targetMolecule = [premises[0][0], 0]
                    experiments.append([targetFileName, sourceMolecule, targetMolecule])
            elif serializedPremise[0][1] == "exclusion":
                if serializedPremise[1][1] == "nullrequirement":
                    targetFileName = serializedPremise[0][0]
                    sourceMolecule = [premises[0][0], 1]
                    targetMolecule = [premises[0][1], 0]
                    experiments.append([targetFileName, sourceMolecule, targetMolecule])

        return experiments

    def lookupPatterns(
        self, graph, center, context, molecule, componentNameIntersection
    ):
        counter = 0
        relevantRules = set([])
        while "{0}%{1}".format(molecule, counter) in graph:
            for component in graph["{0}%{1}".format(molecule, counter)]:
                if center[0] in component:
                    for component2 in graph["{0}%{1}".format(molecule, counter)][
                        component
                    ]:
                        if component2 == context[0]:
                            for activationStatus in graph[
                                "{0}%{1}".format(molecule, counter)
                            ][component][component2]:
                                if context[1] == 1 and (
                                    activationStatus[0] == 1
                                    or activationStatus[1] not in ["", "0"]
                                ):
                                    relevantRules = relevantRules.union(
                                        set(
                                            graph["{0}%{1}".format(molecule, counter)][
                                                component
                                            ][component2][activationStatus]
                                        )
                                    )
                                elif context[1] == 0 and (
                                    activationStatus[0] == 0
                                    and activationStatus[1] in ["", "0"]
                                ):
                                    relevantRules = relevantRules.union(
                                        set(
                                            graph["{0}%{1}".format(molecule, counter)][
                                                component
                                            ][component2][activationStatus]
                                        )
                                    )

            counter += 1
        return relevantRules

    def convertExperimentNames(
        self, componentNameIntersection, molecule, experiment, firstFileFlag
    ):
        trueExperiment = [experiment[0]]

        for element in experiment[1:]:
            if not firstFileFlag:
                trueExperiment.append(element)
            else:
                for component in componentNameIntersection[molecule]:
                    if componentNameIntersection[molecule][component] == element[0]:
                        trueExperiment.append([component, element[1]])
                        break

        return trueExperiment

    def obtainContextDifferences(
        self,
        fileName1,
        fileName2,
        moleculeNameIntersection,
        componentNameIntersection,
        context1,
        context2,
    ):
        (
            file1Context,
            relationshipGraph1,
        ) = context1  # componentGroups.getContextRequirements(fileName1,False)
        (
            file2Context,
            relationshipGraph2,
        ) = context2  # componentGroups.getContextRequirements(fileName2,False)

        # contextMatrix1 = {molecule:[[0]*len(componentNameIntersection[molecule]) for x in componentNameIntersection[molecule]]  for molecule in moleculeNameIntersection}
        contextMatrix1 = prettyDict(lambda: prettyDict(lambda: prettyDict(dict)))
        # contextMatrix2 = prettyDict(lambda: prettyDict(lambda: prettyDict(list)))

        for key in moleculeNameIntersection:
            if (
                key not in file1Context
                or moleculeNameIntersection[key] not in file2Context
            ):
                continue
            for relationship in file1Context[key]:
                for element in file1Context[key][relationship]:
                    if relationship == "independent":
                        continue
                        # moleculePair = [element[0][0],element[1]]
                    else:
                        moleculePair = [element[0][0], element[1][0]]
                        # contextMatrix[element[0]][element[1]] = relationship
                        # continue

                    if (
                        moleculePair[0].lower()
                        in componentNameIntersection[moleculeNameIntersection[key]]
                        and moleculePair[1].lower()
                        in componentNameIntersection[moleculeNameIntersection[key]]
                    ):
                        contextMatrix1[moleculeNameIntersection[key]][
                            componentNameIntersection[moleculeNameIntersection[key]][
                                moleculePair[1].lower()
                            ]
                        ][
                            componentNameIntersection[moleculeNameIntersection[key]][
                                moleculePair[0].lower()
                            ]
                        ][
                            fileName1
                        ] = relationship
            for relationship in file2Context[moleculeNameIntersection[key]]:
                for element in file2Context[moleculeNameIntersection[key]][
                    relationship
                ]:
                    if relationship == "independent":
                        continue
                        # moleculePair = [element[0][0],element[1]]
                    else:
                        moleculePair = [element[0][0], element[1][0]]

                    if moleculePair[0].lower() in [
                        componentNameIntersection[moleculeNameIntersection[key]][x]
                        for x in componentNameIntersection[
                            moleculeNameIntersection[key]
                        ]
                    ] and moleculePair[1].lower() in [
                        componentNameIntersection[moleculeNameIntersection[key]][x]
                        for x in componentNameIntersection[
                            moleculeNameIntersection[key]
                        ]
                    ]:
                        contextMatrix1[moleculeNameIntersection[key]][
                            moleculePair[1].lower()
                        ][moleculePair[0].lower()][fileName2] = relationship

        differences, similCounter, totCounter = self.discoverDifferences(
            contextMatrix1, fileName1, fileName2
        )
        relationshipGraphDict = {
            fileName1: relationshipGraph1,
            fileName2: relationshipGraph2,
        }
        fileContextDict = {fileName1: file1Context, fileName2: file2Context}

        for molecule in differences:
            for componentPair in differences[molecule]:
                counter = 0
                hypothesis = self.hypothesisGenerator(
                    componentPair, fileName1, fileName2
                )
                experimentResults = prettyDict(list)
                for experiment in hypothesis:
                    trueExperiment = self.convertExperimentNames(
                        componentNameIntersection,
                        molecule,
                        experiment,
                        experiment[0] == fileName1,
                    )
                    rules = self.lookupPatterns(
                        relationshipGraphDict[trueExperiment[0]],
                        trueExperiment[1],
                        trueExperiment[2],
                        molecule,
                        componentNameIntersection,
                    )
                    experimentResults[trueExperiment[0]].append(
                        [trueExperiment[1:], rules]
                    )
                componentPair.append(experimentResults)
        return differences, similCounter, totCounter

    def evaluateDifferences(
        self, differences, fileName1, fileName2, fileName1raw, fileName2raw
    ):
        for molecule in differences:
            for differenceInstance in differences[molecule]:
                pass


def compareModelsContext(fileName1, fileName2, context1, context2):
    modelComparison = ModelComparison()

    (
        moleculeNameIntersection,
        componentNameIntersection,
    ) = modelComparison.compareModelsStructurally(
        fileName1, fileName2, moleculemapping, componentmapping
    )
    differences, simCounter, totCounter = modelComparison.obtainContextDifferences(
        fileName1,
        fileName2,
        moleculeNameIntersection,
        componentNameIntersection,
        context1,
        context2,
    )

    # evaluateDifferences(differences, fileName1, fileName2, fileName1raw, fileName2raw)
    # with open(outputle, 'w') as f:
    #    f.write(yaml.dump(differences))
    return fileName1, fileName2, simCounter, totCounter, differences


def defineConsole():
    """
    defines the program console line commands
    """
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-f1", "--file1", type=str, help="reference file", required=True
    )
    parser.add_argument(
        "-f2", "--file2", type=str, help="comparison file", required=True
    )
    # parser.add_argument('-o', '--output', type=str, help='output file', required=True)
    return parser


if __name__ == "__main__":
    parser = defineConsole()
    namespace = parser.parse_args()
    context1 = componentGroups.getContextRequirements(namespace.file1, False)
    context2 = componentGroups.getContextRequirements(namespace.file2, False)

    result = compareModelsContext(namespace.file1, namespace.file2, context1, context2)
    import pprint

    print(pprint.pprint(result))



import os, bionetgen, glob
from tempfile import TemporaryDirectory

from bionetgen.core.utils.logging import BNGLogger


class VisResult:
    """
    Class that's used to load and write in `.graphml` files generated by BioNetGen.

    Helper class for BNGVisualize
    """

    def __init__(self, input_folder, name=None, vtype=None, app=None) -> None:
        self.app = app
        self.logger = BNGLogger(app=self.app)
        self.logger.debug(
            "Setting up VisResult object", loc=f"{__file__} : VisResult.__init__()"
        )
        self.input_folder = input_folder
        self.name = name
        self.vtype = vtype
        self.rc = None
        self.out = None
        self.files = []
        self.file_strs = {}
        self.file_graphs = {}
        self._load_files()

    def _load_files(self) -> None:
        self.logger.debug(
            "Loading graphml/gml files", loc=f"{__file__} : VisResult._load_files()"
        )
        # we need to assume some sort of GML output
        # at least for now
        # use the name, if given, search for GMLs if not
        gmls = glob.glob("*.gml")
        graphmls = glob.glob("*.graphml")
        graphfiles = gmls + graphmls
        for gfile in graphfiles:
            if self.name is None:
                self.files.append(gfile)
                # now load into string
                with open(gfile, "r") as f:
                    l = f.read()
                self.file_strs[gfile] = l
            else:
                # pull GMLs that contain the name
                if self.name in gfile:
                    self.files.append(gfile)
                    # now load into string
                    with open(gfile, "r") as f:
                        l = f.read()
                    self.file_strs[gfile] = l

    def _dump_files(self, folder) -> None:
        self.logger.debug(
            "Writing graphml/gml files", loc=f"{__file__} : VisResult._dump_files()"
        )
        os.chdir(folder)
        for gfile in self.files:
            g_name = os.path.split(gfile)[-1]
            with open(g_name, "w") as f:
                f.write(self.file_strs[gfile])


class BNGVisualize:
    """
    Class that generates various grapsh in `.graphml` format for a model.

    Usage: BNGVisualize(INPUT_FILE, output=OUTPATH, vtype="contactmap", bngpath=BNGPATH)


    Arguments
    ---------
    input_file : str
        path to input `.bngl` file
    output : str
        path to output folder
    vtype: str
        type of visualization. valid options are: "contactmap",
        "ruleviz_pattern", "ruleviz_operation", "regulatory" and "all"
    bngpath : str
        path to BNG2.pl
    suppress : bool
        suppresses any output from BNG2.pl calls
    app : Cement App
        Cement framework app object if working with cement frame work

    Methods
    -------
    run : None
        runs the commands necessary to generate the graph files
    """

    def __init__(
        self, input_file, output=None, vtype=None, bngpath=None, suppress=None, app=None
    ) -> None:
        self.app = app
        self.logger = BNGLogger(app=self.app)
        self.logger.debug(
            "Setting up BNGVisualize object",
            loc=f"{__file__} : BNGVisualize.__init__()",
        )
        # set input, required
        self.input = input_file
        # set valid types
        self.valid_types = [
            "contactmap",
            "ruleviz_pattern",
            "ruleviz_operation",
            "regulatory",
        ]
        self.accept_types = [
            "contactmap",
            "ruleviz_pattern",
            "ruleviz_operation",
            "regulatory",
            "atom_rule",
            "all",
        ]
        # set visualization type, default yo contactmap
        if vtype is None or len(vtype) == 0:
            vtype = "contactmap"
        if vtype not in self.accept_types:
            raise ValueError(f"{vtype} is not a valid visualization type")

        self.vtype = vtype
        # set output
        self.output = output
        self.suppress = suppress
        self.bngpath = bngpath

    def run(self) -> VisResult:
        self.logger.debug("Running", loc=f"{__file__} : BNGVisualize.run()")
        return self._normal_mode()

    def _normal_mode(self):
        self.logger.debug(
            f"Running on normal mode, loading model {self.input}",
            loc=f"{__file__} : BNGVisualize._normal_mode()",
        )
        model = bionetgen.modelapi.bngmodel(self.input)
        model.actions.clear_actions()
        if self.vtype == "all":
            for valid_type in self.valid_types:
                model.add_action("visualize", action_args={"type": f"'{valid_type}'"})
        else:
            if self.vtype == "atom_rule":
                model.add_action("visualize", action_args={"type": f"'regulatory'"})
            else:
                model.add_action("visualize", action_args={"type": f"'{self.vtype}'"})
        # TODO: Work in temp folder
        cur_dir = os.getcwd()
        from bionetgen.core.main import BNGCLI

        self.logger.debug(
            "Generating visualization files",
            loc=f"{__file__} : BNGVisualize._normal_mode()",
        )

        if self.output is None:
            with TemporaryDirectory() as out:
                # instantiate a CLI object with the info
                cli = BNGCLI(model, out, self.bngpath, suppress=self.suppress)
                try:
                    cli.run()
                    # load vis
                    vis_res = VisResult(
                        os.path.abspath(os.getcwd()),
                        name=model.model_name,
                        vtype=self.vtype,
                    )
                    # go back
                    os.chdir(cur_dir)
                    # dump files
                    vis_res._dump_files(cur_dir)
                    return vis_res
                except Exception as e:
                    os.chdir(cur_dir)
                    print("Couldn't run the simulation, see error.")
                    raise e
        else:
            # instantiate a CLI object with the info
            cli = BNGCLI(model, self.output, self.bngpath, suppress=self.suppress)
            try:
                cli.run()
                # load vis
                vis_res = VisResult(
                    os.path.abspath(os.getcwd()),
                    name=model.model_name,
                    vtype=self.vtype,
                )
                # go back
                os.chdir(cur_dir)
                return vis_res
            except Exception as e:
                self.logger.error(
                    "Failed to run file",
                    loc=f"{__file__} : BNGVisualize._normal_mode()",
                )
                os.chdir(cur_dir)
                print("Couldn't run the simulation, see error.")
                raise e



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 14 18:17:20 2014

@author: proto
"""

import libsbml
from util import logMess
from sbml2bngl import SBML2BNGL as SBML2BNGL
import structures
import atomizer.resolveSCT as mc
import os
from subprocess import call
import tempfile
import sys

# sys.path.insert(0, '../utils/')
import consoleCommands
import readBNGXML
import argparse

bioqual = [
    "BQB_IS",
    "BQB_HAS_PART",
    "BQB_IS_PART_OF",
    "BQB_IS_VERSION_OF",
    "BQB_HAS_VERSION",
    "BQB_IS_HOMOLOG_TO",
    "BQB_IS_DESCRIBED_BY",
    "BQB_IS_ENCODED_BY",
    "BQB_ENCODES",
    "BQB_OCCURS_IN",
    "BQB_HAS_PROPERTY",
    "BQB_IS_PROPERTY_OF",
    "BQB_HAS_TAXON",
    "BQB_UNKNOWN",
]

modqual = [
    "BQM_IS",
    "BQM_IS_DESCRIBED_BY",
    "BQM_IS_DERIVED_FROM",
    "BQM_IS_INSTANCE_OF",
    "BQM_HAS_INSTANCE",
    "BQM_UNKNOWN",
]


import fnmatch
import argparse


def getFiles(directory, extension):
    """
    Gets a list of <*.extension> files. include subdirectories and return the absolute
    path. also sorts by size.
    """
    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, "*.{0}".format(extension)):
            matches.append(
                [
                    os.path.join(os.path.abspath(root), filename),
                    os.path.getsize(os.path.join(root, filename)),
                ]
            )

    # sort by size
    matches.sort(key=lambda filename: filename[1], reverse=False)

    matches = [x[0] for x in matches]

    return matches


from collections import defaultdict
import re


def standardizeName(name):
    """
    Remove stuff not used by bngl
    """
    name2 = name

    sbml2BnglTranslationDict = {
        "^": "",
        "'": "",
        "*": "m",
        " ": "_",
        "#": "sh",
        ":": "_",
        "α": "a",
        "β": "b",
        "γ": "g",
        " ": "",
        "+": "pl",
        "/": "_",
        ":": "_",
        "-": "_",
        ".": "_",
        "?": "unkn",
        ",": "_",
        "(": "",
        ")": "",
        "[": "",
        "]": "",
        # "(": "__",
        # ")": "__",
        # "[": "__",
        # "]": "__",
        "(": "",
        ")": "",
        "[": "",
        "]": "",
        ">": "_",
        "<": "_",
    }

    for element in sbml2BnglTranslationDict:
        name = name.replace(element, sbml2BnglTranslationDict[element])
    name = re.sub("[\W]", "", name)
    return name


def parseAnnotation(annotation):
    speciesAnnotationDict = defaultdict(list)
    lista = libsbml.CVTermList()
    libsbml.RDFAnnotationParser.parseRDFAnnotation(annotation, lista)
    for idx in range(0, lista.getSize()):
        for idx2 in range(0, lista.get(idx).getResources().getLength()):
            resource = lista.get(idx).getResources().getValue(idx2)

            qualifierType = lista.get(idx).getQualifierType()
            qualifierDescription = (
                bioqual[lista.get(idx).getBiologicalQualifierType()]
                if qualifierType
                else modqual[lista.get(idx).getModelQualifierType()]
            )
            speciesAnnotationDict[qualifierDescription].append(resource)
    return speciesAnnotationDict


def buildAnnotationDict(document):
    annotationDict = defaultdict(lambda: defaultdict(list))
    speciesNameDict = {}
    for species in document.getModel().getListOfSpecies():
        annotation = species.getAnnotation()
        transformedName = standardizeName(species.getName())
        speciesNameDict[transformedName] = species.getName()
        speciesNameDict[species.getName()] = transformedName
        if annotation:
            annotationDict[transformedName] = parseAnnotation(annotation)
    return annotationDict, speciesNameDict


def updateFromParent(child, parent, annotationDict):
    for annotationLabel in annotationDict[parent]:
        if annotationLabel in [
            "BQB_IS_VERSION_OF",
            "BQB_IS",
            "BQB_IS_HOMOLOG_TO",
            "BQB_HAS_VERSION",
        ]:
            annotationDict[child]["BQB_HAS_VERSION"] = annotationDict[parent][
                annotationLabel
            ]
        elif annotationLabel in ["BQB_HAS_PART"]:
            annotationDict[child][annotationLabel] = annotationDict[parent][
                annotationLabel
            ]


def updateFromChild(parent, child, annotationDict):
    for annotationLabel in annotationDict[child]:
        if annotationLabel in [
            "BQB_IS_VERSION_OF",
            "BQB_IS",
            "BQB_HAS_VERSION",
            "BQB_IS_HOMOLOG_TO",
        ]:
            annotationDict[parent]["BQB_HAS_VERSION"] = annotationDict[child][
                annotationLabel
            ]


# IS_HOMOLOG_TO
def updateFromComplex(complexMolecule, sct, annotationDict, annotationToSpeciesDict):
    localSpeciesDict = {}
    unmatchedReactants = []
    unmatchedAnnotations = []
    for constituentElement in sct[complexMolecule][0]:
        flag = False
        if len(annotationDict[constituentElement]) > 0:
            for annotation in annotationDict[constituentElement]:
                if annotation in [
                    "BQB_IS_VERSION_OF",
                    "BQB_IS",
                    "BQB_HAS_VERSION",
                    "BQB_IS_HOMOLOG_TO",
                    "BQM_IS",
                ]:
                    flag = True
                    for individualAnnotation in annotationDict[constituentElement][
                        annotation
                    ]:
                        localSpeciesDict[individualAnnotation] = constituentElement
                        localSpeciesDict[constituentElement] = individualAnnotation
            if flag:
                continue

        if constituentElement in annotationToSpeciesDict:
            localSpeciesDict[constituentElement] = annotationToSpeciesDict[
                constituentElement
            ]
            localSpeciesDict[
                annotationToSpeciesDict[constituentElement]
            ] = constituentElement
        else:
            unmatchedReactants.append(constituentElement)

    for annotationType in annotationDict[complexMolecule]:
        if annotationType in ["BQB_HAS_VERSION", "BQB_HAS_PART"]:
            for constituentAnnotation in annotationDict[complexMolecule][
                annotationType
            ]:
                if constituentAnnotation not in localSpeciesDict:
                    unmatchedAnnotations.append(constituentAnnotation)
    if len(set(unmatchedReactants)) == 1 and len(set(unmatchedAnnotations)) == 1:
        localSpeciesDict[unmatchedReactants[0]] = unmatchedAnnotations[0]
        localSpeciesDict[unmatchedAnnotations[0]] = unmatchedReactants[0]
        annotationDict[unmatchedReactants[0]]["BQB_IS_VERSION_OF"] = [
            unmatchedAnnotations[0]
        ]

    elif len(unmatchedReactants) > 0 or len(unmatchedAnnotations) > 0:
        # annotate from database names
        print("**//", complexMolecule, unmatchedReactants, unmatchedAnnotations)

    for element in localSpeciesDict:
        if element not in annotationToSpeciesDict:
            annotationToSpeciesDict[element] = localSpeciesDict[element]


def updateFromComponents(complexMolecule, sct, annotationDict, annotationToSpeciesDict):
    localSpeciesDict = defaultdict(set)
    unmatchedReactants = []
    for constituentElement in sct[complexMolecule][0]:
        flag = False
        if complexMolecule == "G_sub_q_endsub__alpha__beta__gamma_":
            print(constituentElement)

        if len(annotationDict[constituentElement]) > 0:
            if complexMolecule == "G_sub_q_endsub__alpha__beta__gamma_":
                print(constituentElement, annotationDict[constituentElement])

            for annotation in annotationDict[constituentElement]:
                if annotation in [
                    "BQB_IS_VERSION_OF",
                    "BQB_IS",
                    "BQB_HAS_VERSION",
                    "BQB_HAS_PART",
                    "BQB_IS_HOMOLOG_TO",
                    "BQM_IS",
                ]:
                    for individualAnnotation in annotationDict[constituentElement][
                        annotation
                    ]:
                        # localSpeciesDict[individualAnnotation] = constituentElement
                        localSpeciesDict[constituentElement].add(individualAnnotation)
                        flag = True
        if not flag:
            unmatchedReactants.append(constituentElement)

    for element in localSpeciesDict:
        annotationDict[complexMolecule]["BQB_HAS_PART"].extend(
            list(localSpeciesDict[element])
        )


def buildAnnotationTree(annotationDict, sct, database):
    annotationToSpeciesDict = {}
    for element in database.weights:
        if len(sct[element[0]]) > 0:
            if len(sct[element[0]][0]) == 1:
                buildingBlock = sct[element[0]][0][0]
                if len(annotationDict[element[0]]) == 0:
                    if len(annotationDict[buildingBlock]) > 0:
                        updateFromParent(element[0], buildingBlock, annotationDict)
                if len(annotationDict[buildingBlock]) == 0:
                    if len(annotationDict[element[0]]) > 0:
                        updateFromChild(buildingBlock, element[0], annotationDict)
            elif len(sct[element[0]][0]) > 1:
                if len(annotationDict[element[0]]) == 0:
                    updateFromComponents(
                        element[0], sct, annotationDict, annotationToSpeciesDict
                    )
                else:
                    if (
                        "BQB_HAS_VERSION" in annotationDict[element[0]]
                        or "BQB_HAS_PART" in annotationDict[element[0]]
                    ):
                        updateFromComplex(
                            element[0], sct, annotationDict, annotationToSpeciesDict
                        )
                    else:
                        updateFromComponents(
                            element[0], sct, annotationDict, annotationToSpeciesDict
                        )
            #        annotationdict[element[0]]
            # for buildingBlock in sct[element[0]][0]:
            #    print '\t',buildingBlock,annotationDict[buildingBlock]


def speciesAnnotationsToSBML(sbmlDocument, annotationDict, speciesNameDict):
    """
    Receives a series of annotations associated with their associated species
    and fills in a corresponding sbmlDocument with this information
    """
    for species in sbmlDocument.getModel().getListOfSpecies():
        transformedName = speciesNameDict[species.getName()]
        if len(annotationDict[transformedName]) == 0:
            continue
        for element in annotationDict[transformedName]:

            term = libsbml.CVTerm()
            if element.startswith("BQB"):
                term.setQualifierType(libsbml.BIOLOGICAL_QUALIFIER)
                term.setBiologicalQualifierType(bioqual.index(element))
            else:
                term.setQualifierType(libsbml.MODEL_QUALIFIER)
                term.setModelQualifierType(modqual.index(element))
            for annotation in annotationDict[transformedName][element]:
                term.addResource(annotation)
            species.addCVTerm(term)

        annotation = libsbml.RDFAnnotationParser.createAnnotation()
        cvterms = libsbml.RDFAnnotationParser.createCVTerms(species)
        rdfAnnotation = libsbml.RDFAnnotationParser.createRDFAnnotation()
        if cvterms:
            rdfAnnotation.addChild(cvterms)
        else:
            print(species)
        annotation.addChild(rdfAnnotation)
        species.setAnnotation(annotation)


actionSboDictionary = {
    "StateChange": "http://identifiers.org/biomodels.sbo/SBO:0000464",
    "DeleteBond": "http://identifiers.org/biomodels.sbo/SBO:0000180",
    "AddBond": "http://identifiers.org/biomodels.sbo/SBO:0000342",
    "ChangeCompartment": "http://identifiers.org/biomodels.sbo/SBO:0000185",
}


def buildReactionAnnotationDict(rules):
    sboDict = defaultdict(lambda: defaultdict(list))
    for rule in rules:
        actions = [x.action for x in rule[0].actions]
        if "Add" not in actions and "Delete" not in actions:
            sboDict[rule[0].label]["BQB_IS_VERSION_OF"] = [
                actionSboDictionary[x] for x in set(actions)
            ]
    return sboDict


def reactionAnnotationsToSBML(sbmlDocument, annotationDict):
    """
    Receives a series of annotations associated with their associated species
    and fills in a corresponding sbmlDocument with this information
    """
    for reaction in sbmlDocument.getModel().getListOfReactions():
        transformedName = reaction.getName()
        if len(annotationDict[transformedName]) == 0:
            continue
        for element in annotationDict[transformedName]:
            term = libsbml.CVTerm()
            if element.startswith("BQB"):
                term.setQualifierType(libsbml.BIOLOGICAL_QUALIFIER)
                term.setBiologicalQualifierType(bioqual.index(element))
            else:
                term.setQualifierType(libsbml.MODEL_QUALIFIER)
                term.setModelQualifierType(modqual.index(element))
            for annotation in annotationDict[transformedName][element]:
                term.addResource(annotation)
            reaction.addCVTerm(term)

        annotation = libsbml.RDFAnnotationParser.createAnnotation()
        cvterms = libsbml.RDFAnnotationParser.createCVTerms(reaction)
        rdfAnnotation = libsbml.RDFAnnotationParser.createRDFAnnotation()
        rdfAnnotation.addChild(cvterms)
        annotation.addChild(rdfAnnotation)
        reaction.setAnnotation(annotation)


def obtainSCT(fileName, reactionDefinitions, useID, namingConventions):
    """
    one of the library's main entry methods. Process data from a file
    to obtain the species composition table, a dictionary describing
    the chemical history of different elements in the system
    """
    logMess.log = []
    logMess.counter = -1
    reader = libsbml.SBMLReader()
    document = reader.readSBMLFromFile(fileName)

    parser = SBML2BNGL(document.getModel(), useID)

    database = structures.Databases()
    database.forceModificationFlag = True
    database.softConstraints = True
    database.parser = parser
    database = mc.createSpeciesCompositionGraph(
        parser,
        database,
        reactionDefinitions,
        namingConventions,
        speciesEquivalences=None,
        bioGridFlag=False,
    )

    return database.prunnedDependencyGraph, database, document, parser.speciesDictionary


import tempfile


def writeSBML(document, fileName):
    writer = libsbml.SBMLWriter()
    writer.writeSBMLToFile(document, fileName)


def createDataStructures(bnglContent):
    """
    create an atomized biomodels in a temporary file to obtain relevant
    bng information
    """

    pointer = tempfile.mkstemp(suffix=".bngl", text=True)
    with open(pointer[1], "w") as f:
        f.write(bnglContent)
    retval = os.getcwd()
    os.chdir(tempfile.tempdir)
    consoleCommands.bngl2xml(pointer[1])
    xmlfilename = ".".join(pointer[1].split(".")[0:-1]) + ".xml"
    os.chdir(retval)
    return readBNGXML.parseXML(xmlfilename)


def expandAnnotation(fileName, bnglFile):

    sct, database, sbmlDocument, _ = obtainSCT(
        fileName,
        "config/reactionDefinitions.json",
        False,
        "config/namingConventions.json",
    )
    annotationDict, speciesNameDict = buildAnnotationDict(sbmlDocument)
    buildAnnotationTree(annotationDict, sct, database)
    speciesAnnotationsToSBML(sbmlDocument, annotationDict, speciesNameDict)
    # species, rules, par = createDataStructures(bnglFile)
    # reactionAnnotationDict = buildReactionAnnotationDict(rules)
    # reactionAnnotationsToSBML(sbmlDocument,reactionAnnotationDict)
    # reactionAnnotationsToSBML(sbmlDocument,reactionAnnotation)

    # reactionAnnotationsToSBML(sbmlDocument)
    writer = libsbml.SBMLWriter()
    return writer.writeSBMLToString(sbmlDocument)


import progressbar


def batchExtensionProcess(directory, outputDir):
    testFiles = getFiles(directory, "xml")
    progress = progressbar.ProgressBar()

    targetFiles = getFiles(outputDir, "xml")
    for fileIdx in progress(range(len(testFiles))):
        file = testFiles[fileIdx]
        if file in [
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000223.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000488.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000293.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000472.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000255.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000424.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000439.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000416.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000182.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000161.xml",
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsRemoved2/BIOMD0000000504.xml",
        ]:
            continue
        if (
            "/home/proto/workspace/RuleWorld/atomizer/SBMLparser/annotationsExpanded2/{0}".format(
                file.split("/")[-1]
            )
            in targetFiles
        ):
            continue
        print(file)
        sbmlInfo = expandAnnotation(file, "")
        outputFile = os.path.join(outputDir, file.split("/")[-1])

        with open(outputFile, "w") as f:
            f.write(sbmlInfo)


def defineConsole():
    parser = argparse.ArgumentParser(description="SBML to BNGL translator")
    parser.add_argument(
        "-i", "--input-file", type=str, help="input SBML file", required=True
    )
    parser.add_argument(
        "-o", "--output-file", type=str, help="output SBML file", required=True
    )
    return parser


if __name__ == "__main__":
    batchExtensionProcess("annotationsRemoved2", "annotationsExpanded2")

    # parser = defineConsole()
    # namespace = parser.parse_args()
    # input_file = '/home/proto/workspace/bionetgen/parsers/SBMLparser/XMLExamples/curated/BIOMD%010i.xml' % 19
    # expandedString = expandAnnotation(namespace.input_file, '')
    # print 'Writing extended annotation SBML to {0}'.format(namespace.output_file)
    # with open(namespace.output_file,'w') as f:
    #     f.write(expandedString)
    # outputFileName = '.'.join(fileName.split('.')[0:-1]) + '_withAnnotations.xml'
    # writeSBML(sbmlDocument,outputFileName)



class BNGSimulator:
    """
    The generic simulator interface to all BNG simulators.

    This is (going to be) designed to allow for a generic API to all
    other types of simulators so that this API can be used in the rest
    of the library without specifically writing code for each simulator
    type.


    Attributes
    ----------
    model_file : str
        path to the model file this simulator is using

    Properties
    ----------
    simulator: obj
        the python object that runs the actual simulation.
        Any subclass implementing a BNGSimulator should set
        this property and make it so that setting the simulator
        object to the file path given to initialize the
        BNGSimulator class initializes the simulator object in turn.

    Methods
    -------
    simulate(args)
        Uses the arguments provided to call the underlying simulator
    """

    def __init__(self, model_file=None, model_str=None):
        """
        This class is inialized by giving it the path to the model file
        you want it to open and run.
        """
        # this also sets up simulator attribute
        if model_file is not None:
            self.model_file = model_file
        if model_str is not None:
            self.model_str = model_str

    @property
    def model_file(self):
        """
        model file attribute that stores the path to the original
        model file and also sets up the simulator
        """
        return self._model_file

    @model_file.setter
    def model_file(self, mfile):
        self._model_file = mfile
        self.simulator = mfile

    @property
    def model_str(self):
        """
        model file attribute that stores the path to the original
        model file and also sets up the simulator
        """
        return self._model_str

    @model_str.setter
    def model_str(self, mstr):
        self._model_str = mstr
        self.simulator = mstr

    def simulate(self):
        raise NotImplementedError



import ctypes, os, tempfile, bionetgen
import numpy as np

from distutils import ccompiler
from .bngsimulator import BNGSimulator
from bionetgen.main import BioNetGen
from bionetgen.core.exc import BNGCompileError

# This allows access to the CLIs config setup
app = BioNetGen()
app.setup()
conf = app.config["bionetgen"]
def_bng_path = conf["bngpath"]


class RESULT(ctypes.Structure):
    _fields_ = [
        ("status", ctypes.c_int),
        ("n_observables", ctypes.c_int),
        ("n_species", ctypes.c_int),
        ("n_tpts", ctypes.c_int),
        ("obs_name_len", ctypes.c_int),
        ("spcs_name_len", ctypes.c_int),
        ("observables", ctypes.POINTER(ctypes.c_double)),
        ("species", ctypes.POINTER(ctypes.c_double)),
        ("obs_names", ctypes.POINTER(ctypes.c_char)),
        ("spcs_names", ctypes.POINTER(ctypes.c_char)),
    ]


class CSimWrapper:
    """
    Wrapper class for the compiled C simulator shared library.

    The class loads the compiled C shared library and passes
    pointers to the initial species arrays and parameter arrays
    to the shared library, runs the simulation and returns the
    results as numpy named arrays.
    """

    def __init__(self, lib_path, num_params=None, num_spec_init=None):
        # we need the result struct to reconstruct the object
        self.return_struct = RESULT
        # load the shared library
        self.lib = ctypes.CDLL(lib_path)
        # set the return type of the simulate function to a pointer
        # we'll use it to reconstruct the RESULT object from it later
        self.lib.simulate.restype = ctypes.c_void_p
        # set number of parameters
        self.num_params = num_params
        # set number of initial species values
        self.num_spec_init = num_spec_init

    def set_species_init(self, arr):
        """
        Set the initial species values array
        """
        # TODO: Transition to BNGErrors and logging
        assert len(arr) == self.num_spec_init
        self.species_init = np.array(arr, dtype=np.float64)

    def set_parameters(self, arr):
        """
        Set the parameter values array
        """
        # TODO: Transition to BNGErrors and logging
        assert len(arr) == self.num_params
        self.parameters = np.array(arr, dtype=np.float64)

    def simulate(self, t_start=0, t_end=100, n_steps=100):
        """
        Run the simulate command of the shared C library.

        This function will construct a timepoint array from the given
        arguments and then pass the timepoints, parameters and
        initial species to the simulate command. Take the result pointer
        and convert the pointer back to a result struct and then
        construct named numpy arrays to return observable and species
        values over time.
        """
        # generate the time point array
        del_t = (t_end - t_start) / float(n_steps)
        timepoints = np.arange(t_start, t_end + 1, del_t)
        ntpts = len(timepoints)
        # call the simulate command
        self.result = self.return_struct.from_address(
            self.lib.simulate(
                ntpts,
                ctypes.c_void_p(timepoints.ctypes.data),
                self.num_spec_init,
                ctypes.c_void_p(self.species_init.ctypes.data),
                self.num_params,
                ctypes.c_void_p(self.parameters.ctypes.data),
            )
        )
        # we need to pull the observable names and get a list of them
        obs_names = ctypes.cast(
            self.result.obs_names,
            ctypes.POINTER(ctypes.c_char * self.result.obs_name_len),
        )[0].value.decode()
        obs_names = obs_names.split("/")[:-1]
        # same thing with species names
        spcs_names = ctypes.cast(
            self.result.spcs_names,
            ctypes.POINTER(ctypes.c_char * self.result.spcs_name_len),
        )[0].value.decode()
        spcs_names = spcs_names.split("/")[:-1]
        # cast the observables into a named numpy array
        buffer_as_ctypes_arr_obs = ctypes.cast(
            self.result.observables,
            ctypes.POINTER(ctypes.c_double * ntpts * self.result.n_observables),
        )[0]
        observables = np.frombuffer(buffer_as_ctypes_arr_obs, np.float64)
        fmt = ["f8"] * len(obs_names)
        obs_all = np.reshape(observables, (self.result.n_observables, ntpts))
        obs_all = np.core.records.fromarrays(obs_all, names=obs_names, formats=fmt)
        # cast the species into a named numpy array
        buffer_as_ctypes_arr_spc = ctypes.cast(
            self.result.species,
            ctypes.POINTER(ctypes.c_double * ntpts * self.result.n_species),
        )[0]
        species = np.frombuffer(buffer_as_ctypes_arr_spc, np.float64)
        fmt = ["f8"] * len(spcs_names)
        spcs_all = np.reshape(species, (self.result.n_species, ntpts))
        spcs_all = np.core.records.fromarrays(spcs_all, names=spcs_names, formats=fmt)
        # free the memory used for the results struct
        self.lib.free_result(ctypes.byref(self.result))
        del self.result
        # return named numpy arrays
        return (obs_all, spcs_all)


class CSimulator(BNGSimulator):
    """
    Object that bridges the BNG model object and the CSimWrapper object.

    The point of this object is to deal with the compilation of the shared library
    and pass the correct parameter and initial species values to the wrapper object.
    """

    def __init__(self, model_file, generate_network=False):
        # check cvode library paths
        if (conf.get("cvode_include") is None) or (conf.get("cvode_lib") is None):
            print("CVODE include and library paths are not set, compilation won't work")
        # let's load the model first
        if isinstance(model_file, str):
            # load model file
            self.model = bionetgen.bngmodel(
                model_file, generate_network=generate_network
            )
        elif isinstance(model_file, bionetgen.bngmodel):
            # loaded model
            self.model = model_file
            cd = os.getcwd()
            with tempfile.TemporaryDirectory() as tmpdirname:
                os.chdir(tmpdirname)
                self.model.write_model(f"{self.model.model_name}.bngl")
                self.model = bionetgen.bngmodel(
                    f"{self.model.model_name}.bngl", generate_network=generate_network
                )
            os.chdir(cd)
        else:
            print(f"model format not recognized: {model_file}")
        # set compiler
        self.compiler = ccompiler.new_compiler()
        self.compiler.add_include_dir(conf.get("cvode_include"))
        self.compiler.add_library_dir(conf.get("cvode_lib"))
        # compile shared library
        self.compile_shared_lib()
        # setup simulator
        self.simulator = self.lib_file

    def __str__(self):
        return f"C/Python Simulator, params: {self.model.parameters} \ninit species: {self.model.species}"

    def __repr__(self):
        return str(self)

    def compile_shared_lib(self):
        # run and get CPY file
        # make sure we don't have actions
        self.model.actions.clear_actions()
        self.model.actions.add_action("generate_network", {"overwrite": 1})
        self.model.actions.add_action("writeCPYfile", {})
        # for now run and write the .c file in the current folder
        bionetgen.run(self.model, out=os.path.abspath(os.getcwd()))
        # compile CPY file
        c_file = f"{self.model.model_name}_cvode_py.c"
        obj_file = f"{self.model.model_name}_cvode_py.o"
        lib_file = f"{self.model.model_name}_cvode_py"
        # compile objects with fPIC for the shared lib we'll link
        self.compiler.compile([c_file], extra_preargs=["-fPIC"])
        # now link cvode and nvecserial and make a shared lib
        self.compiler.link_shared_lib(
            [obj_file], lib_file, libraries=["sundials_cvode", "sundials_nvecserial"]
        )
        # # keep a record of what we got
        self.cfile = os.path.abspath(c_file)
        self.obj_file = os.path.abspath(obj_file)
        # compiler tacks on the lib at the beginning and .so at the end
        lib_file = f"lib{self.model.model_name}_cvode_py.so"
        self.lib_file = os.path.abspath(lib_file)

    @property
    def simulator(self):
        """
        simulator attribute that stores
        the instantiated simulator object
        and also saves the SBML text in the
        sbml attribute
        """
        return self._simulator

    @simulator.setter
    def simulator(self, lib_file):
        # use CSimWrapper under the hood
        try:
            n_param = len(
                list(
                    filter(lambda x: not x.startswith("_"), list(self.model.parameters))
                )
            )
            self._simulator = CSimWrapper(
                os.path.abspath(lib_file),
                num_params=n_param,
                num_spec_init=len(self.model.species),
            )
        except:
            raise BNGCompileError(self.model)

    def simulate(self, t_start=0, t_end=10, n_steps=10):
        # set parameters and initial species values
        spcs = []
        for spc_name in self.model.species:
            try:
                count = float(self.model.species[spc_name].count)
                spcs.append(count)
            except:
                p_name = self.model.species[spc_name].count
                count = float(self.model.parameters[p_name].value)
                spcs.append(count)
        self.simulator.set_species_init(spcs)
        params = list(filter(lambda x: not x.startswith("_"), self.model.parameters))
        params = [float(self.model.parameters[p].value) for p in params]
        self.simulator.set_parameters(params)
        # now that we have CSimWrapper setup correctly, run the simulation
        obs_all, spcs_all = self.simulator.simulate(t_start, t_end, n_steps)
        # return our results
        return (obs_all, spcs_all)
